---
created_at: '2016-07-31T14:37:42.000Z'
title: Lisp as an Alternative to Java (1999)
url: http://norvig.com/java-lisp.html
author: nwhitfield1
points: 77
story_text: 
comment_text: 
num_comments: 103
story_id: 
story_title: 
story_url: 
parent_id: 
created_at_i: 1469975862
_tags:
- story
- author_nwhitfield1
- story_12197131
objectID: '12197131'

---
[Source](http://norvig.com/java-lisp.html "Permalink to Lisp as an Alternative to Java")

# Lisp as an Alternative to Java

# Lisp as an Alternative to Java

In the October 1999 _Communications of the ACM_ [Lutz Prechelt][1] had an interesting article entitled [Comparing Java vs. C/C++ Efficiency Issues to Interpersonal Issues][2] which asked 38 programmers to implement versions of a program in C, C++, or Java. The conclusions showed that Java was 3 or 4 times slower than C or C++, but that the variance between programmers was larger than the variance between languages, suggesting that one might want to spend more time on training programmers rather than arguing over language choice. (Or, suggesting that you should hire the good programmers and avoid the bad ones.) The variance for Java was lower than for C or C++. (Cynics could say that Java forces you to write uniformly slow programs.) I applaud this line of work, and hope that more studies of this kind will be done. 

It turns out my hopes were answered. First, Prechelt published [another article][3] that covers Tcl, Python, Perl, and Rexx. Also, [Ron Garret (nee Erann Gat)][4] did a [ follow-up study][5] in which he asked programmers to write Prechelt's [test program][6] in Lisp. His results show that the resulting Lisp programs ran faster on average than C, C++ or Java programs (although the fastest Lisp program was not as fast as the fastest C program), and that Lisp programs took less development time than the other languages. 

I did not participate in the study, but after I saw it, I wrote my version in Lisp. It took me about 2 hours (compared to a range of 2 to 8.5 hours for the other Lisp programmers in the study, 3 to 25 for C/C++ and 4 to 63 for Java) and I ended up with 45 non-comment non-blank lines (compared with a range of 51 to 182 for Lisp, and 107 to 614 for the other languages). (That means that some Java programmer was spending 13 lines and 84 minutes to provide the functionality of each line of my Lisp program.) 

Here is my program: 

| ----- |
| 
    
    
    ;; Peter Norvig - Programming Challange from Erann Gat:
    ;; <http://www.flownet.com/ron/papers/lisp-java/>
    ;; Given a list of words and a list of phone numbers, find all the ways that
    ;; each phone number can be expressed as a list of words.
    ;; Run: (main "word-list-file-name" "phone-number-file-name")
    
    (defvar ***dict*** nil 
      "A hash table mapping a phone number (integer) to a list of words from the 
      input dictionary that produce that number.")
    
    (defun **main** (&optional; (dict "dict") (nums "nums") (dict-size 100))
      "Read the input file DICT and load it into *dict*.  Then for each line in 
      NUMS, print all the translations of the number into a sequence of words,
      according to the rules of translation."
      (setf *dict* (load-dictionary dict dict-size))
      (with-open-file (in nums)
        (loop for num = (read-line in nil) while num do
              (print-translations num (remove-if-not #'digit-char-p num)))))
    
    (defun **print-translations** (num digits &optional; (start 0) (words nil))
      "Print each possible translation of NUM into a string of words.  DIGITS
      must be WORD with non-digits removed.  On recursive calls, START is the
      position in DIGITS at which to look for the next word, and WORDS is the list
      of words found for (subseq DIGITS 0 START).  So if START gets to the end of
      DIGITS, then we have a solution in WORDS.  Otherwise, for every prefix of
      DIGITS, look in the dictionary for word(s) that map to the value of the
      prefix (computed incrementally as N), and for each such word try to extend
      the solution with a recursive call.  There are two complications: (1) the
      rules say that in addition to dictionary words, you can use a single 
      digit in the output, but not two digits in a row. Also (and this seems 
      silly) you can't have a digit in a place where any word could appear.
      I handle this with the variable FOUND-WORD; if it is false after the loop, 
      and the most recent word is not a digit, try a recursive call that pushes a 
      digit. (2) The other complication is that the obvious way of mapping 
      strings to integers would map R to 2 and ER to 02, which of course is 
      the same integer as 2.  Therefore we prepend a 1 to every number, and R 
      becomes 12 and ER becomes 102."
      (if (>= start (length digits))
          (format t "~a:~{ ~a~}~%" num (reverse words))
          (let ((found-word nil)
                (n 1)) ; leading zero problem
            (loop for i from start below (length digits) do
                  (setf n (+ (* 10 n) (nth-digit digits i)))
                  (loop for word in (gethash n *dict*) do
                     (setf found-word t)
                     (print-translations num digits (+ 1 i) (cons word words))))
            (when (and (not found-word) (not (numberp (first words))))
               (print-translations num digits (+ start 1) 
                                   (cons (nth-digit digits start) words))))))
    
    (defun **load-dictionary** (file size)
      "Create a hashtable from the file of words (one per line).  Takes a hint
      for the initial hashtable size.  Each key is the phone number for a word;
      each value is a list of words with that phone number."
      (let ((table (make-hash-table :test #'eql :size size)))
        (with-open-file (in file)
          (loop for word = (read-line in nil) while word do
            (push word (gethash (word->number word) table))))
        table))
    
    (defun **word->number** (word)
      "Translate a word (string) into a phone number, according to the rules."
      (let ((n 1)) ; leading zero problem
        (loop for i from 0 below (length word) 
              for ch = (char word i) do
              (when (alpha-char-p ch) (setf n (+ (* 10 n) (char->digit ch)))))
        n))
    
    (defun **nth-digit** (digits i) 
      "The i-th element of a character string of digits, as an integer 0 to 9."
      (- (char-code (char digits i)) #.(char-code #
    

[1]: http://wwwipd.ira.uka.de/~prechelt/
[2]: http://www.acm.org/pubs/articles/journals/cacm/1999-42-10/p109-prechelt/p109-prechelt.pdf
[3]: http://wwwipd.ira.uka.de/~prechelt/Biblio/jccpprtTR.pdf
[4]: http://www.flownet.com/ron/
[5]: http://www.flownet.com/gat/papers/lisp-java.pdf
[6]: http://www.flownet.com/ron/papers/lisp-java/

